"""Module to generate overall activity summary from epoch data."""

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset

from actinet.utils.utils import date_parser, toScreen
from actinet import circadian

ACTIVITY_LABELS = ["light", "moderate-vigorous", "sedentary", "sleep"]


def getActivitySummary(
    data,
    intensityDistribution=False,
    circadianMetrics=False,
    verbose=True,
):
    """
    Calculate overall activity summary from predicted activity label data.
    This is achieved by:
    1) check if data occurs at a daylight savings crossover
    2) calculate imputation values to replace nan PA metric values
    3) calculate empirical cumulative distribution function of vector magnitudes
    4) derive main movement summaries (overall, weekday/weekend, and hour)

    :param str data: Input csv.gz file or pandas dataframe of processed epoch data
    :param bool intensityDistribution: Add intensity outputs to dict <summary>
    :param bool circadianMetrics: Add circadian rhythm metrics to dict <summary>
    :param bool verbose: Print verbose output

    :return: A summary of the activity.
    :rtype: dict

    """

    toScreen("=== Summarizing ===", verbose)

    if isinstance(data, str):
        data = pd.read_csv(
            data,
            index_col=["time"],
            parse_dates=["time"],
            date_parser=date_parser,
        )

    # Main movement summaries
    summary = _summarise(
        data, ACTIVITY_LABELS, intensityDistribution, circadianMetrics, verbose, summary
    )

    # Return physical activity summary
    return summary


def _summarise(
    data,
    labels,
    intensityDistribution=False,
    circadianMetrics=False,
    verbose=False,
    summary={},
):
    """Overall summary stats for each activity type to summary dict

    :param pandas.DataFrame data: Pandas dataframe of epoch data
    :param list(str) labels: Activity state labels
    :param dict summary: Output dictionary containing all summary metrics
    :param bool intensityDistribution: Add intensity outputs to dict <summary>
    :param bool circadianMetrics: Add circadian rhythm metrics to dict <summary>
    :param bool verbose: Print verbose output
    :param dict summary: Output dictionary containing all summary metrics

    :return: Updated dict <summary> keys for each activity type 'overall-<avg/sd>',
        'week<day/end>-avg', '<day..>-avg', 'hourOfDay-<hr..>-avg',
        'hourOfWeek<day/end>-<hr..>-avg'
    :rtype: dict
    """

    data = data.copy()
    freq = to_offset(pd.infer_freq(data.index))

    # Get start day
    startTime = data.index[0]
    summary["FirstDay(0=mon,6=sun)"] = startTime.weekday()

    # Check daylight savings crossings
    summary = checkDST(data, summary)

    # Hours of activity for each recorded day
    epochPeriod = int(pd.Timedelta(freq).total_seconds())
    cols = labels
    dailyStats = (
        data[cols].astype("float").groupby(data.index.date).sum() * epochPeriod / 3600
    ).reset_index(drop=True)

    for i, row in dailyStats.iterrows():
        for col in cols:
            summary[f"day{i}-recorded-{col}(hrs)"] = row.loc[col]

    # Calculate empirical cumulative distribution function of vector magnitudes
    if intensityDistribution:
        summary = calculateECDF(data["acc"], summary)

    # In the following, we resample, pad and impute the data so that we have a
    # multiple of 24h for the stats calculations
    tStart, tEnd = data.index[0], data.index[-1]
    cols = ["acc"] + labels
    if "MET" in data.columns:
        cols.append("MET")
    data = imputeMissing(data[cols].astype("float"))

    # Overall stats (no padding, i.e. only within recording period)
    toScreen("=== Calculating overall statistics ===", verbose)
    overallStats = data[tStart:tEnd].apply(["mean", "std"])
    for col in overallStats:
        summary[f"{col}-overall-avg"] = overallStats[col].loc["mean"]
        summary[f"{col}-overall-sd"] = overallStats[col].loc["std"]

    dayOfWeekStats = data.groupby([data.index.weekday, data.index.hour]).mean()
    dayOfWeekStats.index = dayOfWeekStats.index.set_levels(
        dayOfWeekStats.index.levels[0]
        .to_series()
        .replace({0: "mon", 1: "tue", 2: "wed", 3: "thu", 4: "fri", 5: "sat", 6: "sun"})
        .to_list(),
        level=0,
    )
    dayOfWeekStats.index.set_names(["DayOfWeek", "Hour"], inplace=True)

    # Week stats
    for col, value in dayOfWeekStats.mean().items():
        summary[f"{col}-week-avg"] = value

    # Stats by day of week (Mon, Tue, ...)
    for col, stats in dayOfWeekStats.groupby(level=0).mean().to_dict().items():
        for dayOfWeek, value in stats.items():
            summary[f"{col}-{dayOfWeek}-avg"] = value

    # Stats by hour of day
    for col, stats in dayOfWeekStats.groupby(level=1).mean().to_dict().items():
        for hour, value in stats.items():
            summary[f"{col}-hourOfDay-{hour}-avg"] = value

    weekdayOrWeekendStats = dayOfWeekStats.groupby(
        [
            dayOfWeekStats.index.get_level_values("DayOfWeek").str.contains("sat|sun"),
            dayOfWeekStats.index.get_level_values("Hour"),
        ]
    ).mean()
    weekdayOrWeekendStats.index = weekdayOrWeekendStats.index.set_levels(
        weekdayOrWeekendStats.index.levels[0]
        .to_series()
        .replace({True: "Weekend", False: "Weekday"})
        .to_list(),
        level=0,
    )
    weekdayOrWeekendStats.index.set_names(["WeekdayOrWeekend", "Hour"], inplace=True)

    # Weekday/weekend stats
    for col, stats in weekdayOrWeekendStats.groupby(level=0).mean().to_dict().items():
        for weekdayOrWeekend, value in stats.items():
            summary[f"{col}-{weekdayOrWeekend.lower()}-avg"] = value

    # Stats by hour of day AND by weekday/weekend
    for col, stats in weekdayOrWeekendStats.to_dict().items():
        for key, value in stats.items():
            weekdayOrWeekend, hour = key
            summary[f"{col}-hourOf{weekdayOrWeekend}-{hour}-avg"] = value

    # Calculate circadian metrics
    if circadianMetrics:
        toScreen("=== Calculating circadian metrics ===", verbose)
        summary = circadian.calculatePSD(
            data, epochPeriod, False, ACTIVITY_LABELS, summary
        )
        summary = circadian.calculatePSD(
            data, epochPeriod, True, ACTIVITY_LABELS, summary
        )
        summary = circadian.calculateFourierFreq(
            data, epochPeriod, False, ACTIVITY_LABELS, summary
        )
        summary = circadian.calculateFourierFreq(
            data, epochPeriod, False, ACTIVITY_LABELS, summary
        )
        summary = circadian.calculateM10L5(data, epochPeriod, summary)

    return summary


def checkDST(data, summary={}):
    if data.index[0].dst() < data.index[-1].dst():
        summary["quality-daylightSavingsCrossover"] = 1
    elif data.index[0].dst() > data.index[-1].dst():
        summary["quality-daylightSavingsCrossover"] = -1
    else:
        summary["quality-daylightSavingsCrossover"] = 0

    return summary


def imputeMissing(data, extrapolate=True):
    """Impute missing/nonwear segments

    Impute non-wear data segments using the average of similar time-of-day values
    with one minute granularity on different days of the measurement. This
    imputation accounts for potential wear time diurnal bias where, for example,
    if the device was systematically less worn during sleep in an individual,
    the crude average vector magnitude during wear time would be a biased
    overestimate of the true average. See
    https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0169649#sec013

    :param pandas.DataFrame e: Pandas dataframe of epoch data
    :param bool verbose: Print verbose output

    :return: Update DataFrame <e> columns nan values with time-of-day imputation
    :rtype: void
    """

    if extrapolate:
        # padding at the boundaries to have full 24h
        data = data.reindex(
            pd.date_range(
                data.index[0].floor("D"),
                data.index[-1].ceil("D"),
                freq=to_offset(pd.infer_freq(data.index)),
                closed="left",
                name="time",
            ),
            method="nearest",
            tolerance=pd.Timedelta("1m"),
            limit=1,
        )

    def fillna(subframe):
        # Transform will first pass the subframe column-by-column as a Series.
        # After passing all columns, it will pass the entire subframe again as a DataFrame.
        # Processing the entire subframe is optional (return value can be omitted). See 'Notes' in transform doc.
        if isinstance(subframe, pd.Series):
            x = subframe.to_numpy()
            nan = np.isnan(x)
            nanlen = len(x[nan])
            if 0 < nanlen < len(x):  # check x contains a NaN and is not all NaN
                x[nan] = np.nanmean(x)
                return x  # will be cast back to a Series automatically
            else:
                return subframe

    data = (
        data
        # first attempt imputation using same day of week
        .groupby([data.index.weekday, data.index.hour, data.index.minute])
        .transform(fillna)
        # then try within weekday/weekend
        .groupby([data.index.weekday >= 5, data.index.hour, data.index.minute])
        .transform(fillna)
        # finally, use all other days
        .groupby([data.index.hour, data.index.minute])
        .transform(fillna)
    )

    return data


def calculateECDF(x, summary):
    """Calculate activity intensity empirical cumulative distribution

    The input data must not be imputed, as ECDF requires different imputation
    where nan/non-wear data segments are IMPUTED FOR EACH INTENSITY LEVEL. Here,
    the average of similar time-of-day values is imputed with one minute
    granularity on different days of the measurement. Following intensity levels
    are calculated:
    1mg bins from 1-20mg
    5mg bins from 25-100mg
    25mg bins from 125-500mg
    100mg bins from 500-2000mg

    :param pandas.DataFrame e: Pandas dataframe of epoch data
    :param str inputCol: Column to calculate intensity distribution on
    :param dict summary: Output dictionary containing all summary metrics

    :return: Updated summary file
    :rtype: dict
    """

    levels = np.concatenate(
        [
            np.linspace(1, 20, 20),  # 1mg bins from 1-20mg
            np.linspace(25, 100, 16),  # 5mg bins from 25-100mg
            np.linspace(125, 500, 16),  # 25mg bins from 125-500mg
            np.linspace(600, 2000, 15),  # 100mg bins from 500-2000mg
        ]
    ).astype("int")

    whrnan = x.isna().to_numpy()
    ecdf = x.to_numpy().reshape(-1, 1) <= levels.reshape(1, -1)
    ecdf[whrnan] = np.nan

    ecdf = (
        pd.DataFrame(ecdf, index=x.index, columns=levels)
        .groupby([x.index.hour, x.index.minute])
        .mean()  # first average is across same time of day
        .mean()  # second average is within each level
    )

    # Write to summary
    for level, val in ecdf.iteritems():
        summary[f"{x.name}-ecdf-{level}mg"] = val

    return summary



import argparse
import os
import json
from collections import OrderedDict
import pandas as pd
from tqdm.auto import tqdm


def collate_outputs(outputs, outfile="outputs.csv"):
    """Read all *-outputSummary.json files under <outputs> and merge into one CSV file.
    :param str outputs: Directory containing JSON files.
    :param str outfile: Output CSV filename.
    :return: New file written to <outfile>
    :rtype: void
    """

    # Load all *-outputSummary.json files under outputs/
    infofiles = []
    jdicts = []
    for root, _, files in os.walk(outputs):
        for file in files:
            if file.endswith("-outputSummary.json"):
                infofiles.append(os.path.join(root, file))

    print(f"Found {len(infofiles)} summary files...")
    for file in tqdm(infofiles):
        with open(file, "r") as f:
            jdicts.append(json.load(f, object_pairs_hook=OrderedDict))

    summary = pd.DataFrame.from_dict(jdicts)  # merge to a dataframe
    summary.to_csv(outfile, index=False)
    print("Summary CSV written to", outfile)

    return


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("outputs")
    parser.add_argument("--outfile", "-o", default="outputs.csv")
    args = parser.parse_args()

    return collate_outputs(outputs=args.outputs, outfile=args.outfile)


if __name__ == "__main__":
    main()



import datetime
import re
import numpy as np
import pandas as pd
from scipy import stats


def infer_freq(x):
    """Like pd.infer_freq but more forgiving"""
    freq, _ = stats.mode(np.diff(x), keepdims=False)
    freq = pd.Timedelta(freq)
    return freq


def toScreen(msg, verbose=True):
    """
    Print msg str prepended with current time

    :param str mgs: Message to be printed to screen
    :return: None. Prints msg str prepended with current time.
    """
    if verbose:
        timeFormat = "%Y-%m-%d %H:%M:%S"
        print(f"\n{datetime.datetime.now().strftime(timeFormat)}\t{msg}")


def date_parser(t):
    """
    Parse date a date string of the form e.g.
    2020-06-14 19:01:15.123+0100 [Europe/London]
    """
    tz = re.search(r"(?<=\[).+?(?=\])", t)
    if tz is not None:
        tz = tz.group()
    t = re.sub(r"\[(.*?)\]", "", t)
    return pd.to_datetime(t, utc=True).tz_convert(tz)



import numpy as np
import os
import pandas as pd


class HMM:
    """
    Implement a basic HMM model with parameter saving/loading.
    """

    def __init__(
        self,
        prior=None,
        emission=None,
        transition=None,
        labels=None,
        uniform_prior=True,
    ):
        self.prior = prior
        self.emission = emission
        self.transition = transition
        self.labels = labels
        self.uniform_prior = uniform_prior

    def __str__(self):
        return (
            "Hidden Markov Model\n"
            "prior: {prior}\n"
            "emission: {emission}\n"
            "transition: {transition}\n"
            "labels: {labels}".format(
                prior=self.prior,
                emission=self.emission,
                transition=self.transition,
                labels=self.labels,
            )
        )

    def train(self, y_prob, y_true, t=None, interval=None):
        """https://en.wikipedia.org/wiki/Hidden_Markov_model
        :param y_prob: Observation probabilities
        :param y_true: Ground truth labels
        """

        if self.labels is None:
            self.labels = np.unique(y_true)

        prior = np.mean(y_true.reshape(-1, 1) == self.labels, axis=0)

        emission = np.vstack(
            [np.mean(y_prob[y_true == label], axis=0) for label in self.labels]
        )

        transition = calculate_transition_matrix(y_true, t, interval)

        self.prior = prior
        self.emission = emission
        self.transition = transition

    def predict(self, y_obs, t=None, interval=None, uniform_prior=None):
        check_for_time_values_error(y_obs, t, interval)

        y_smooth = self.viterbi(y_obs, uniform_prior)

        if t is not None:
            y_smooth = restore_labels_after_gaps(y_obs, y_smooth, t, interval)

        return y_smooth

    def viterbi(self, y_obs, uniform_prior=None):
        """Perform HMM smoothing over observations via Viteri algorithm
        https://en.wikipedia.org/wiki/Viterbi_algorithm
        :param y_obs: Predicted observation
        :param bool uniform_prior: Assume uniform priors.

        :return: Smoothed sequence of activities
        :rtype: np.ndarray
        """

        def log(x):
            return np.log(x + 1e-16)

        prior = (
            np.ones(len(self.labels)) / len(self.labels)
            if (self.uniform_prior or uniform_prior)
            else self.prior
        )
        emission = self.emission
        transition = self.transition
        labels = self.labels

        nobs = len(y_obs)
        n_labels = len(labels)

        y_obs = np.where(y_obs.reshape(-1, 1) == labels)[1]  # to numeric

        probs = np.zeros((nobs, n_labels))
        probs[0, :] = log(prior) + log(emission[:, y_obs[0]])
        for j in range(1, nobs):
            for i in range(n_labels):
                probs[j, i] = np.max(
                    log(emission[i, y_obs[j]]) + log(transition[:, i]) + probs[j - 1, :]
                )  # probs already in log scale
        viterbi_path = np.zeros_like(y_obs)
        viterbi_path[-1] = np.argmax(probs[-1, :])
        for j in reversed(range(nobs - 1)):
            viterbi_path[j] = np.argmax(
                log(transition[:, viterbi_path[j + 1]]) + probs[j, :]
            )  # probs already in log scale

        viterbi_path = labels[viterbi_path]  # to labels

        return viterbi_path

    def save(self, path):
        """
        Save model parameters to a Numpy npz file.

        :param str path: npz file location
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        np.savez(
            path,
            prior=self.prior,
            emission=self.emission,
            transition=self.transition,
            labels=self.labels,
        )

    def load(self, path):
        """
        Load model parameters from a Numpy npz file.

        :param str path: npz file location
        """
        d = np.load(path, allow_pickle=True)
        self.prior = d["prior"]
        self.emission = d["emission"]
        self.transition = d["transition"]
        self.labels = d["labels"]


def check_for_time_values_error(Y, T, interval):
    # If truthy, T must be the same length as Y, and interval must also be truthy
    if T is not None:
        if len(Y) != len(T):
            raise Exception("Provided times should have same length as labels")
        if not interval:
            raise Exception(
                "A window length must be provided when using label times to train hmm"
            )


def restore_labels_after_gaps(y_pred, y_smooth, t, interval):
    # Restore unsmoothed predictions to labels following gaps in time
    df = pd.DataFrame({"y_pred": y_pred, "y_smooth": y_smooth})

    if type(t[0]) == int:
        gaps = pd.Series(t).diff(periods=1) != interval
        gaps[0] = False
    else:
        gaps = pd.Series(t).diff(periods=1) != pd.Timedelta(seconds=interval)
        gaps[0] = False

    df.loc[gaps, "y_smooth"] = df.loc[gaps, "y_pred"]

    return df["y_smooth"].values


def calculate_transition_matrix(Y, t=None, interval=None):
    # t and interval are used to identify any gaps in the data
    # If not provided, it is assumed there are no gaps
    check_for_time_values_error(Y, t, interval)

    t = t if t is not None else range(len(Y))
    interval = interval or 1

    df = pd.DataFrame(Y)

    # create a new column with data shifted one space
    df["shift"] = df[0].shift(-1)

    # only consider transitions of expected interval
    if type(t[0]) == int:
        df = df[(-pd.Series(t).diff(periods=-1) == interval)]
    else:
        df = df[(-pd.Series(t).diff(periods=-1) == pd.Timedelta(seconds=interval))]

    # add a count column (for group by function)
    df["count"] = 1

    # groupby and then unstack, fill the zeros
    trans_mat = df.groupby([0, "shift"]).count().unstack().fillna(0)

    # normalise by occurences and save values to get the transition matrix
    return trans_mat.div(trans_mat.sum(axis=1), axis=0).values



import os
import pathlib
import urllib
import shutil
import time
import argparse
import json
import hashlib
import numpy as np
import pandas as pd
import joblib

import actipy

from actinet import __model_version__
from actinet import __model_md5__
from actinet.accPlot import plotTimeSeries
from actinet.models import ActivityClassifier
from actinet.sslmodel import SAMPLE_RATE
from actinet.summarisation import getActivitySummary, ACTIVITY_LABELS
from actinet.utils.utils import infer_freq

BASE_URL = "https://zenodo.org/records/10619096/files/"


def main():

    parser = argparse.ArgumentParser(
        description="A tool to predict activities from accelerometer data using a self-supervised Resnet 18 model",
        add_help=True,
    )
    parser.add_argument("--filepath", "-f", help="Enter file to be processed")
    parser.add_argument(
        "--outdir",
        "-o",
        help="Enter folder location to save output files",
        default="outputs/",
    )
    parser.add_argument(
        "--model-path", "-m", help="Enter custom model file to use", default=None
    )
    parser.add_argument(
        "--force-download", action="store_true", help="Force download of model file"
    )
    parser.add_argument(
        "--pytorch-device",
        "-d",
        help="Pytorch device to use, e.g.: 'cpu' or 'cuda:0'",
        type=str,
        default="cpu",
    )
    parser.add_argument(
        "--sample-rate",
        "-r",
        help="Sample rate for measurement, otherwise inferred.",
        type=int,
        default=None,
    )
    parser.add_argument(
        "--plot-activity",
        "-p",
        action="store_true",
        help="Plot the predicted activity labels",
    )
    parser.add_argument(
        "--cache-ssl",
        action="store_true",
        help="Download and cache ssl module for offline usage",
    )
    parser.add_argument(
        "--ssl-repo-path", "-s", help="Enter repository of ssl model", default=None
    )
    parser.add_argument("--quiet", "-q", action="store_true", help="Suppress output")
    args = parser.parse_args()

    before = time.time()

    verbose = not args.quiet

    if args.cache_ssl:
        model = ActivityClassifier(weights_path=None, ssl_repo=None, verbose=verbose, labels=ACTIVITY_LABELS)

        after = time.time()
        print(f"Done! ({round(after - before,2)}s)")

        return

    # Load file
    data, info = read(
        args.filepath,
        resample_hz=SAMPLE_RATE,
        sample_rate=args.sample_rate,
        verbose=verbose,
    )

    # Output paths
    basename = resolve_path(args.filepath)[1]
    outdir = os.path.join(args.outdir, basename)
    os.makedirs(outdir, exist_ok=True)

    # Run model
    if verbose:
        print("Loading model...")
    model_path = pathlib.Path(__file__).parent / f"{__model_version__}.joblib.lzma"
    check_md5 = args.model_path is None
    model: ActivityClassifier = load_model(
        args.model_path or model_path, check_md5, args.force_download, verbose
    )

    model.verbose = verbose
    model.device = args.pytorch_device

    if verbose:
        print("Running activity classifier...")
    Y = model.predict_from_frame(data)

    # Save predicted activities
    timeSeriesFile = f"{outdir}/{basename}-timeSeries.csv.gz"
    Y.to_csv(timeSeriesFile)

    if verbose:
        print("Time series output written to:", timeSeriesFile)

    # Summary
    summary = getActivitySummary(Y, True, True, verbose)

    # Join the actipy processing info, with acitivity summary data
    outputSummary = {**summary, **info}

    # Save output summary
    outputSummaryFile = f"{outdir}/{basename}-outputSummary.json"
    with open(outputSummaryFile, "w") as f:
        json.dump(outputSummary, f, indent=4, cls=NpEncoder)

    if verbose:
        print("Output summary written to:", outputSummaryFile)

    # Print
    if verbose:
        print("\nSummary Stats\n---------------------")
        print(
            {
                key: outputSummary[key]
                for key in [
                    "Filename",
                    "Filesize(MB)",
                    "WearTime(days)",
                    "NonwearTime(days)",
                    "ReadOK",
                    "acc-overall-avg(mg)",
                ]
                + [f"{label}-week-avg" for label in ACTIVITY_LABELS]
            }
        )

    # Plot activity profile
    if args.plot_activity:
        plotFile = f"{outdir}/{basename}-timeSeries-plot.png"
        fig = plotTimeSeries(Y)
        fig.savefig(plotFile, dpi=200, bbox_inches="tight")

        if verbose:
            print("Output plot written to:", plotFile)

    after = time.time()
    print(f"Done! ({round(after - before,2)}s)")


def read(filepath, resample_hz="uniform", sample_rate=None, verbose=True):

    p = pathlib.Path(filepath)
    ftype = p.suffixes[0].lower()
    fsize = round(p.stat().st_size / (1024 * 1024), 1)

    if ftype in (".csv", ".pkl"):

        if ftype == ".csv":
            data = pd.read_csv(
                filepath,
                usecols=["time", "x", "y", "z"],
                parse_dates=["time"],
                index_col="time",
                dtype={"x": "f4", "y": "f4", "z": "f4"},
            )
        elif ftype == ".pkl":
            data = pd.read_pickle(filepath)
        else:
            raise ValueError(f"Unknown file format: {ftype}")

        if sample_rate in (None, False):
            freq = infer_freq(data.index)
            sample_rate = int(np.round(pd.Timedelta("1s") / freq))

        # Quick fix: Drop duplicate indices. TODO: Maybe should be handled by actipy.
        data = data[~data.index.duplicated(keep="first")]

        data, info = actipy.process(
            data,
            sample_rate,
            lowpass_hz=None,
            calibrate_gravity=True,
            detect_nonwear=True,
            resample_hz=resample_hz,
            verbose=verbose,
        )

        info = {
            **{
                "Filename": filepath,
                "Device": ftype,
                "Filesize(MB)": fsize,
                "SampleRate": sample_rate,
            },
            **info,
        }

    elif ftype in (".cwa", ".gt3x", ".bin"):

        data, info = actipy.read_device(
            filepath,
            lowpass_hz=None,
            calibrate_gravity=True,
            detect_nonwear=True,
            resample_hz=resample_hz,
            verbose=verbose,
        )

    if "ResampleRate" not in info:
        info["ResampleRate"] = info["SampleRate"]

    return data, info


def resolve_path(path):
    """Return parent folder, file name and file extension"""
    p = pathlib.Path(path)
    extension = p.suffixes[0]
    filename = p.name.rsplit(extension)[0]
    dirname = p.parent
    return dirname, filename, extension


def load_model(model_path, ssl_repo_path=None, check_md5=True, force_download=False, verbose=True):
    """Load trained model. Download if not exists."""

    pth = pathlib.Path(model_path)

    if force_download or not pth.exists():

        url = f"{BASE_URL}{__model_version__}.joblib.lzma"

        if verbose:
            print(f"Downloading {url}...")

        with urllib.request.urlopen(url) as f_src, open(pth, "wb") as f_dst:
            shutil.copyfileobj(f_src, f_dst)

    if check_md5:
        assert md5(pth) == __model_md5__, (
            "Model file is corrupted. Please run with --force-download "
            "to download the model file again."
        )

    model: ActivityClassifier = joblib.load(pth)

    if ssl_repo_path and pathlib.Path(ssl_repo_path).exists():
        if verbose:
            print(f"Loading ssl repository from {ssl_repo_path}.")

        model = model.load_ssl(ssl_repo_path)

    return model


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
from tqdm.auto import tqdm
from torch.utils.data import DataLoader

from actinet import hmm
from actinet import sslmodel


class ActivityClassifier:
    def __init__(
        self,
        device="cpu",
        batch_size=512,
        window_sec=30,
        weights_path=None,
        labels=[],
        ssl_repo=None,
        repo_tag="v1.0.0",
        hmm_params=None,
        verbose=False,
    ):
        self.device = device
        self.repo_tag = repo_tag
        self.batch_size = batch_size
        self.window_sec = window_sec
        self.labels = labels
        self.window_len = int(np.ceil(self.window_sec * sslmodel.SAMPLE_RATE))
        self.verbose = verbose
        
        self.model_weights = sslmodel.get_model_dict(weights_path, device) if weights_path else None
        self.model = self.load_ssl(ssl_repo)

        hmm_params = hmm_params or dict()
        self.hmms = hmm.HMM(**hmm_params)

    def __str__(self):
        return (
            "Activity Classifier\n"
            "class_labels: {self.labels}\n"
            "window_length: {self.window_sec}\n"
            "batch_size: {self.batch_size}\n"
            "device: {self.device}\n"
            "hmm: {self.hmms}\n"
            "model: {self.model}".format(self=self)
        )

    def predict_from_frame(self, data):

        def fn(chunk):
            """Process the chunk. Apply padding if length is not enough."""
            n = len(chunk)
            x = chunk[["x", "y", "z"]].to_numpy()
            if n == self.window_len:
                x = x
            elif n > self.window_len:
                x = x[: self.window_len]
            elif n < self.window_len and n > self.window_len / 2:
                m = self.window_len - n
                x = np.pad(x, ((0, m), (0, 0)), mode="wrap")
            else:
                x = np.full((self.window_len, 3), fill_value=np.nan)
            return x

        X, T = make_windows(
            data, self.window_sec, fn=fn, return_index=True, verbose=self.verbose
        )

        Y = raw_to_df(X, self._predict(X), T, self.labels, reindex=False)

        return Y

    def _predict(self, X):
        sslmodel.verbose = self.verbose

        dataset = sslmodel.NormalDataset(X)
        dataloader = DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=0,
        )

        _, y_pred, _ = sslmodel.predict(
            self.model, dataloader, self.device, output_logits=False
        )

        y_pred = self.hmms.predict(y_pred)

        return y_pred

    def load_ssl(self, ssl_repo):
        model = sslmodel.get_sslnet(
            tag=self.repo_tag,
            local_repo_path=ssl_repo,
            pretrained_weights = self.model_weights or True,
            window_sec=self.window_sec,
            num_labels=len(self.labels),
        )
        model.to(self.device)

        return model


def make_windows(data, window_sec, fn=None, return_index=False, verbose=True):
    """Split data into windows"""

    if verbose:
        print("Defining windows...")

    if fn is None:

        def fn(x):
            return x

    X, T = [], []
    for t, x in tqdm(
        data.resample(f"{window_sec}s", origin="start"),
        mininterval=5,
        disable=not verbose,
    ):
        x = fn(x)
        X.append(x)
        T.append(t)

    X = np.asarray(X)

    if return_index:
        T = pd.DatetimeIndex(T, name=data.index.name)
        return X, T

    return X


def raw_to_df(data, labels, time, classes, label_proba=False, reindex=True, freq="30S"):
    """
    Construct a DataFrome from the raw data, prediction labels and time Numpy arrays.

    :param data: Numpy windowed acc data, shape (rows, window_len, 3)
    :param labels: Either a scalar label array with shape (rows, ),
                    or the probabilities for each class if label_proba==True with shape (rows, len(classes)).
    :param time: Numpy time array, shape (rows, )
    :param classes: Array with the categorical class labels.
                    The index of this array should correspond to the labels value when label_proba==False.
    :param label_proba: If True, assume 'labels' contains the raw class probabilities.
    :param reindex: Reindex the dataframe to fill missing values
    :param freq: Reindex frequency
    :return: Dataframe
        Index: DatetimeIndex
        Columns: acc, classes
    :rtype: pd.DataFrame
    """
    label_matrix = np.zeros((len(time), len(classes)), dtype=np.float32)
    a_matrix = np.zeros(len(time), dtype=np.float32)

    for i, data in enumerate(data):
        if not label_proba:
            label = int(labels[i])
            label_matrix[i, label] = 1

        x = data[:, 0]
        y = data[:, 1]
        z = data[:, 2]

        enmo = (np.sqrt(x**2 + y**2 + z**2) - 1) * 1000  # in milli gravity
        enmo[enmo < 0] = 0
        a_matrix[i] = np.mean(enmo)

    if label_proba:
        datadict = {
            **{"time": time, "acc": a_matrix},
            **{classes[i]: label_matrix[:, i] for i in range(len(classes))},
        }
    else:
        datadict = {
            **{"time": time, "acc": a_matrix},
            **{classes[i]: label_matrix[:, i] for i in range(len(classes))},
        }

    df = pd.DataFrame(datadict)
    df = df.set_index("time")

    if reindex:
        newindex = pd.date_range(df.index[0], df.index[-1], freq=freq)
        df = df.reindex(newindex, method="nearest", fill_value=np.nan, tolerance="5S")

    return df



"""Script to plot accelerometer traces."""

import sys
import numpy as np
import pandas as pd
from pandas.plotting import register_matplotlib_converters
import os
import matplotlib.patches as mpatches
import matplotlib.lines as mlines
import matplotlib.pyplot as plt
from datetime import datetime, timedelta, time
import argparse
import matplotlib

from actinet.utils import utils

# http://pandas-docs.github.io/pandas-docs-travis/whatsnew/v0.21.1.html#restore-matplotlib-datetime-converter-registration
register_matplotlib_converters()

LABELS_AND_COLORS = {
    "imputed": "#fafc6f",
    "sleep": "midnightblue",
    "sit-stand": "red",
    "sedentary": "red",
    "vehicle": "saddlebrown",
    "light": "darkorange",
    "mixed": "seagreen",
    "walking": "limegreen",
    "moderate-vigorous": "green",
    "moderate": "green",
    "vigorous": "springgreen",
    "bicycling": "springgreen",
    "tasks-light": "darkorange",
    "SB": "red",  # sedentary behaviour
    "LPA": "darkorange",  # light physical activity
    "MVPA": "green",  # moderate-vigorous physical activity
    "MPA": "green",  # moderate physical activity
    "VPA": "springgreen",  # vigorous physical activity
}


def main():  # noqa: C901
    """
    Application entry point responsible for parsing command line requests
    """

    parser = argparse.ArgumentParser(
        description="A script to plot acc time series data.", add_help=True
    )
    # required
    parser.add_argument(
        "timeSeriesFile",
        metavar="input file",
        type=str,
        help="input .csv.gz time series file to plot",
    )
    parser.add_argument(
        "--plotFile",
        metavar="output file",
        type=str,
        help="output .png file to plot to",
    )
    parser.add_argument(
        "--showFileName",
        metavar="True/False",
        default=False,
        type=str2bool,
        help="""Toggle showing filename as title in output
                            image (default : %(default)s)""",
    )
    parser.add_argument(
        "--showFirstNDays",
        metavar="days",
        default=None,
        type=int,
        help="Show just first n days",
    )

    # check input is ok
    if len(sys.argv) < 2:
        msg = "\nInvalid input, please enter at least 1 parameter, e.g."
        msg += "\npython accPlot.py timeSeries.csv.gz \n"
        utils.toScreen(msg)
        parser.print_help()
        sys.exit(-1)
    args = parser.parse_args()

    # determine output file name
    if args.plotFile is None:
        inputFileFolder, inputFileName = os.path.split(args.timeSeriesFile)
        inputFileName = inputFileName.split(".")[0]  # remove any extension
        args.plotFile = os.path.join(inputFileFolder, inputFileName + "-plot.png")

    # read time series file to pandas DataFrame
    data = pd.read_csv(
        args.timeSeriesFile,
        index_col="time",
        parse_dates=["time"],
        date_parser=utils.date_parser,
    )

    # set backend if run from main
    matplotlib.use("Agg")

    # set plot title
    title = args.timeSeriesFile if args.showFileName else None

    # call plot function and save figure
    fig = plotTimeSeries(data, title=title, showFirstNDays=args.showFirstNDays)
    fig.savefig(args.plotFile, dpi=200, bbox_inches="tight")
    print("Plot file written to:", args.plotFile)


def plotTimeSeries(data, title=None, showFirstNDays=None):  # noqa: C901
    """
    Plot acceleration traces and classified activities.

    :param data: Input time-series of acceleration and activity classes. Index: DatetimeIndex. Columns (4-class example):
                 - Name: acc, dtype=float (optional)
                 - Name: light, dtype=Any numeric, value=0 or 1
                 - Name: moderate-vigorous, dtype=Any numeric, value=0 or 1
                 - Name: sedentary, dtype=Any numeric, value=0 or 1
                 - Name: sleep, dtype=Any numeric, value=0 or 1
    :type data: pd.DataFrame
    :param title: Optional plot title
    :type title: str, optional
    :param showFirstNDays: Only show first n days of time series (if specified)
    :type showFirstNDays: int, optional
    :return: pyplot Figure
    :rtype: plt.Figure

    :Example:

    .. code-block:: python

        from accelerometer.accPlot import plotTimeSeries
        df = pd.DataFrame(...)
        fig = plotTimeSeries(df)
        fig.show()
    """

    # check index is datetime
    if not isinstance(data.index, pd.DatetimeIndex):
        raise ValueError("Index must be a DatetimeIndex")

    # use tz-naive local time
    if data.index.tz is not None:
        data.index = (
            # convoluted way to mask the ambiguous DST pushback hour, if any
            data.index.tz_localize(None)
            .tz_localize(data.index.tz, ambiguous="NaT", nonexistent="NaT")
            .tz_localize(None)
        )
        data = data[data.index.notnull()]

    # fix gaps or irregular sampling
    if pd.infer_freq(data) is None:
        freq = (
            pd.infer_freq(data.head()) or "30s"
        )  # try to infer from first few rows, else default to 30s
        new_index = pd.date_range(data.index[0], data.index[-1], freq=freq)
        data = data.reindex(
            new_index, method="nearest", tolerance=freq, fill_value=np.NaN
        )

    if showFirstNDays is not None:
        data = data.first(str(showFirstNDays) + "D")

    labels = [label for label in LABELS_AND_COLORS.keys() if label in data.columns]
    colors = [LABELS_AND_COLORS[label] for label in labels]

    if "imputed" in data.columns:
        mask = data["imputed"].astype("bool")
        labels_excl_imputed = [label for label in labels if label != "imputed"]
        data.loc[mask, labels_excl_imputed] = 0
        data.loc[mask, "acc"] = np.nan

    # setup plotting range
    MAXRANGE = 2 * 1000  # 2g (above this is very rare)
    if "acc" in data:
        data["acc"] = data["acc"].rolling("1T").mean()  # minute average
        data["acc"] = data["acc"].clip(0, MAXRANGE)
    data[labels] = data[labels].astype("f4") * MAXRANGE

    # number of rows to display in figure (all days + legend)
    groupedDays = data.groupby(data.index.date)
    nrows = len(groupedDays) + 1

    # create overall figure
    fig = plt.figure(None, figsize=(10, nrows), dpi=100)
    if title is not None:
        fig.suptitle(title)

    # create individual plot for each day
    i = 0
    axs = []
    for day, group in groupedDays:

        ax = fig.add_subplot(nrows, 1, i + 1)

        if "acc" in group:
            ax.plot(group.index, group["acc"].to_numpy(), c="k")

        if len(labels) > 0:
            ax.stackplot(
                group.index, group[labels].to_numpy().T, colors=colors, edgecolor="none"
            )

        # add date label to left hand side of each day's activity plot
        ax.set_ylabel(
            day.strftime("%A\n%d %B"),
            weight="bold",
            horizontalalignment="right",
            verticalalignment="center",
            rotation="horizontal",
            fontsize="medium",
            color="k",
            labelpad=5,
        )
        # run gridlines for each hour bar
        ax.get_xaxis().grid(True, which="major", color="grey", alpha=0.5)
        ax.get_xaxis().grid(True, which="minor", color="grey", alpha=0.25)
        # set x and y-axes
        ax.set_xlim(group.index[0], group.index[-1])
        ax.set_xticks(
            pd.date_range(
                start=datetime.combine(day, time(0, 0, 0, 0)),
                end=datetime.combine(day + timedelta(days=1), time(0, 0, 0, 0)),
                freq="4H",
            )
        )
        ax.set_xticks(
            pd.date_range(
                start=datetime.combine(day, time(0, 0, 0, 0)),
                end=datetime.combine(day + timedelta(days=1), time(0, 0, 0, 0)),
                freq="1H",
            ),
            minor=True,
        )
        ax.set_ylim(0, MAXRANGE)
        ax.get_yaxis().set_ticks([])  # hide y-axis lables
        # make border less harsh between subplots
        ax.spines["top"].set_color("#d3d3d3")  # lightgray
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_visible(False)
        # set background colour to lightgray
        ax.set_facecolor("#d3d3d3")

        # append to list and incrament list counter
        axs.append(ax)
        i += 1

    # create new subplot to display legends
    ax = fig.add_subplot(nrows, 1, i + 1)
    ax.axis("off")
    legend_patches = [mlines.Line2D([], [], color="k", label="acceleration")]
    for label, color in zip(labels, colors):
        legend_patches.append(mpatches.Patch(color=color, label=label))
    # create overall legend
    plt.legend(
        handles=legend_patches,
        bbox_to_anchor=(0.0, 0.0, 1.0, 1.0),
        loc="center",
        ncol=4,
        mode="best",
        borderaxespad=0,
        framealpha=0.6,
        frameon=True,
        fancybox=True,
    )

    # remove legend border
    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    axs.append(ax)

    # format x-axis to show hours
    fig.autofmt_xdate()
    # add hour labels to top of plot
    hrLabels = ["00:00", "04:00", "08:00", "12:00", "16:00", "20:00", "24:00"]
    axs[0].set_xticklabels(hrLabels)
    axs[0].tick_params(labelbottom=False, labeltop=True, labelleft=False)

    # auto trim borders
    fig.tight_layout()

    return fig


def str2bool(v):
    """
    Used to parse true/false values from the command line. E.g. "True" -> True
    """

    return v.lower() in ("yes", "true", "t", "1")


if __name__ == "__main__":
    main()  # Standard boilerplate to call the main() function to begin the program.



name = "actinet"
__author__ = "Shing Chan, Aidan Acquah, Charilaos Zisou, Aiden Doherty"
__maintainer__ = "Shing Chan"
__maintainer_email__ = "shing.chan@ndph.ox.ac.uk"
__license__ = "See LICENSE file."

__model_version__ = "ssl_ukb_c24_rw_20240204"
__model_md5__ = "84f3d5bb73de5c4da057918c45400da4"

from . import _version

__version__ = _version.get_versions()["version"]
